# -*- coding: utf-8 -*-

"""
    Tests for general and data utils (data_utils.py)
"""

# imports
import cftime
from parameterized import parameterized
import unittest
import numpy as np
from jsmetrics.utils import data_utils
from . import (
    set_up_test_uv_data,
)

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


class TestXarrayDataCheck(unittest.TestCase):
    def setUp(self):
        self.data = set_up_test_uv_data()

    def test_check_coords_in_data(self):
        tested_func = data_utils.check_coords_in_data
        self.assertRaises(KeyError, lambda: tested_func(self.data, ("notthere",)))

    def test_check_variables_in_data(self):
        tested_func = data_utils.check_variables_in_data
        self.assertRaises(
            KeyError,
            lambda: tested_func(self.data, req_variables=["notthere"]),
        )

    def test_check_at_least_n_plevs_in_data(self):
        tested_func = data_utils.check_at_least_n_plevs_in_data
        tested_func(self.data, n_plevs=1)
        self.assertRaises(ValueError, lambda: tested_func(self.data, n_plevs=10))

    def test_check_if_data_is_xarray_datatype(self):
        tested_func = data_utils.check_if_data_is_xarray_datatype
        tested_func(self.data)
        new_data = self.data["ua"].data
        self.assertRaises(TypeError, lambda: tested_func(new_data))


class TestAddNumDaysto360Datetime(unittest.TestCase):
    @parameterized.expand(
        [
            (
                cftime.Datetime360Day(day=1, month=1, year=2000, hour=1),
                360,
                cftime.Datetime360Day(day=1, month=1, year=2001, hour=1),
            ),
            (
                cftime.Datetime360Day(day=1, month=1, year=2000, hour=1),
                30,
                cftime.Datetime360Day(day=1, month=2, year=2000, hour=1),
            ),
            (
                cftime.Datetime360Day(day=1, month=1, year=2000, hour=1),
                29,
                cftime.Datetime360Day(day=30, month=1, year=2000, hour=1),
            ),
            (
                cftime.Datetime360Day(day=1, month=1, year=2000, hour=1),
                359,
                cftime.Datetime360Day(day=30, month=12, year=2000, hour=1),
            ),
        ]
    )
    def test_add_num_days_to360Datetime_expected_vals(
        self, test_date, days_to_add, expected_new_date
    ):
        added_test_date = data_utils.add_num_of_days_to_360Datetime(
            test_date, num_of_days_to_add=days_to_add
        )
        self.assertEqual(added_test_date, expected_new_date)

    def test_errors_raised(self):
        tested_func = data_utils.add_num_of_days_to_360Datetime
        test_date = cftime.Datetime360Day(day=1, month=1, year=2000, hour=1)
        self.assertRaises(
            AssertionError, lambda: tested_func(np.datetime64("1970-01-11"), 1)
        )
        self.assertRaises(
            AssertionError,
            lambda: tested_func(cftime.DatetimeAllLeap(day=1, month=1, year=2000), 1),
        )
        self.assertRaises(ValueError, lambda: tested_func(test_date, 0))
        self.assertRaises(ValueError, lambda: tested_func(test_date, -1))


class TestAddNumDaystoNoLeapDatetime(unittest.TestCase):
    @parameterized.expand(
        [
            (
                cftime.DatetimeNoLeap(day=1, month=1, year=2020, hour=1),
                365,
                cftime.DatetimeNoLeap(day=1, month=1, year=2021, hour=12),
            ),
            (
                cftime.DatetimeNoLeap(day=1, month=1, year=2019, hour=1),
                365,
                cftime.DatetimeNoLeap(day=1, month=1, year=2020, hour=12),
            ),
            (
                cftime.DatetimeNoLeap(day=28, month=2, year=2020, hour=1),
                1,
                cftime.DatetimeNoLeap(day=1, month=3, year=2020, hour=12),
            ),
        ]
    )
    def test_add_num_of_days_to_NoLeapDatetime(
        self, test_date, days_to_add, expected_new_date
    ):
        added_test_date = data_utils.add_num_of_days_to_NoLeapDatetime(
            test_date, num_of_days_to_add=days_to_add
        )
        self.assertEqual(added_test_date, expected_new_date)

    def test_errors_raised(self):
        tested_func = data_utils.add_num_of_days_to_NoLeapDatetime
        test_date = cftime.DatetimeNoLeap(day=1, month=1, year=2000, hour=1)
        self.assertRaises(
            AssertionError, lambda: tested_func(np.datetime64("1970-01-11"), 1)
        )
        self.assertRaises(
            AssertionError,
            lambda: tested_func(cftime.DatetimeAllLeap(day=1, month=1, year=2000), 1),
        )
        self.assertRaises(ValueError, lambda: tested_func(test_date, 0))
        self.assertRaises(ValueError, lambda: tested_func(test_date, -1))


class TestLocalMinimaMaxima(unittest.TestCase):
    def setUp(self):
        self.data = set_up_test_uv_data()

    def test_get_local_maxima(self):
        tested_func = data_utils.get_local_maxima
        self.assertListEqual(
            list(tested_func(self.data["ua"].isel(time=0, plev=0, lon=0).data)[0]),
            [4, 30, 38, 44, 56, 61, 70],
        )

    def test_get_local_minima(self):
        tested_func = data_utils.get_local_minima
        self.assertListEqual(
            list(tested_func(self.data["ua"].isel(time=0, plev=0, lon=0).data)[0]),
            [43, 48, 58, 65],
        )


class TestGetNumOfDecimalPlaces(unittest.TestCase):
    def test_func(self):
        tested_func = data_utils.get_num_of_decimal_places
        self.assertEqual(tested_func(2.33), 2)
        self.assertEqual(tested_func(2), 0)



# -*- coding: utf-8 -*-

"""
    Wrappers for checking inputs to the metrics
"""
import functools
import xarray


def sort_xarray_data_coords(coords, ascending=True):
    """
    Sort the coordinates of the input data (e.g. lat or lon). Sort in ascending order by default.
    Used in all jet statistics and jet core algorithms included in this package.

    Parameters
    ----------
    coords : array-like
        Coords to sort in xarray data
    ascending : Boolean (default=True)
        Whether the coords should be sorted in ascending order
    """

    def wrap(func):
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if "data" in kwargs:
                data = kwargs["data"]
            elif args:
                #  Assumes first argument is data
                data = args[0]
            else:
                return func(*args, **kwargs)
            check_data_is_xarray(data)
            for coord in coords:
                assert (
                    coord in data.coords
                ), f"'{coord}' is not in data. Please check your data and the variable names. Should be like: 'lat', 'lon', 'plev', etc."
                if data[coord].size == 1 and not data[coord].shape == (1,):
                    data = data.expand_dims(
                        coord
                    )  # expand dimensions so the sortby function works if only one in dim
                data = data.sortby(coord, ascending=ascending)
            return func(data, *args[1:], **kwargs)

        return wrapped_func

    return wrap


def check_data_is_xarray(data):
    assert isinstance(data, xarray.DataArray) or isinstance(
        data, xarray.Dataset
    ), "'data' needs to be a xarray DataArray or Dataset type"



# -*- coding: utf-8 -*-

"""
    Methods that return a mask of coordinates related to the jet location, e.g., identifying the maximum
    wind speed throughout the horizontal and/or vertical plane within a given time window.

    The following algorithms each return a xarray.Dataset and are ordered by paper publish year.
"""

# imports
from . import jet_core_algorithms_components
from jsmetrics.utils import windspeed_utils
from jsmetrics.core.check_data import sort_xarray_data_coords

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


@sort_xarray_data_coords(coords=["lat", "lon"])
def koch_et_al_2006(data, ws_threshold=30):
    r"""
    This method follows a two-step procedure used to detect jet-event occurences (here: 'jet_events_ws').

    The weighted average windspeed (:math:`\alpha vel`) for the jet events is calculated as follows:

    .. math::
        \alpha vel =  \frac{1}{p2-p1} \int_{p1}^{p2} (u^2+v^2)^{1/2} \,dp

    where :math:`p1`, :math:`p2` is min, max pressure level.

    After calculating :math:`\alpha vel`, in a second step a windspeed threshold to isolate jet events (the default is :math:`30 m s^{-1}`).

    This method was first introduced in Koch et al (2006) (https://doi.org/10.1002/joc.1255)
    and is described in section 2.2.2 of that study. The original methodology provides a third step
    (to produce a climatology of jet events), but this has been ignored in this implementation.
    Instead, we have provided an example of how to calculate this after running this method
    in 'Examples' below.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    ws_threshold : int or float
        Windspeed threshold used to extract from weighted average (default: 30 ms-1)

    Returns
    ----------
    output : xr.Dataset
        Data containing the output variable 'jet_events_ws'

    Notes
    -----
    The original methodology uses windspeed between 100-400 hPa to calculated the weighted average
    and 30 meters per second as the windspeed threshold.

    This equation for this method is provided on pg 287 of the Koch et al. 2006 paper.

    In the original paper, they accumulate the jet events into two-class jet typology (described in section 2.2.3
    of Koch et al. 2006)

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range used in original methodology (100-400 hPa)):
        uv_sub = uv_data.sel(plev=slice(100, 400))

        # Run algorithm:
        koch_outputs = jsmetrics.jet_core_algorithms.koch_et_al_2006(uv_sub, ws_threshold=30)

        # Produce climatology of jet occurence events for each season and each month:
        koch_month_climatology = koch_outputs.groupby("time.month").mean("time")
        koch_season_climatology = koch_outputs.groupby("time.season").mean("time")

    """
    # Step 1. Check there are at least two plev coordinates for weighted average.
    if data["plev"].count() < 2:
        raise ValueError(
            "Need at least two plevs to calculate weighted average windspeed"
        )

    # Step 2: get all pressure levels (hPa) as list
    all_plevs_hPa = jet_core_algorithms_components.get_all_hPa_list(data)

    # Step 3: get weighted sum windspeed
    sum_weighted_ws = jet_core_algorithms_components.get_sum_weighted_ws(
        data, all_plevs_hPa
    )

    # Step 4: calculate average weighted
    weighted_average_ws = jet_core_algorithms_components.get_weighted_average_ws(
        sum_weighted_ws, all_plevs_hPa
    )

    # Step 5: Apply windspeed threshold to get jet event array
    jet_events = weighted_average_ws.where(weighted_average_ws >= ws_threshold)
    jet_events = jet_events.fillna(0.0)

    # Step 6: Turn jet event array into dataset
    jet_event_ds = jet_events.rename("jet_events_ws").to_dataset()
    return jet_event_ds


@sort_xarray_data_coords(coords=["lat", "lon"])
def schiemann_et_al_2009(data, ws_threshold=30, u_threshold=0):
    r"""
    This method detects 'jet occurrences', whereby each jet occurence is detected based
    on three rules applied to inputted wind speed (:math:`V = [u, v]`):

    1. :math:`|V|` is a local maxima in latitude and altitude plane
    2. :math:`|V| \ge 30 m s^{-1}`
    3. :math:`u \ge 0 m s^{-1}`.

    The implementation of this method here allows you to edit the :math:`|V|` threshold
    (by changing 'ws_threshold'), and :math:`u` threshold (by changing 'u_threshold').

    This method was originally introduce in Schiemann et al 2009 (https://doi.org/10.1175/2008JCLI2625.1)
    and is described in Section 2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    ws_threshold : int or float
        Windspeed threshold used to extract jet maxima from resultant windspeed (default: 30 m/s)
    u_threshold : int or float
        Windspeed threshold used to extract u-component wind speed (default: 0 m/s)

    Returns
    ----------
    output : xr.Dataset
        Data containing the two output variables: 'ws' and 'jet_occurence'

    Notes
    -----
    While the original method is built on a four dimension slice of wind speed (time, lat, lon, plev),
    This implementation will work where there is only one pressure level, so a 3-d slice (time, lat, lon).

    **Slow method:** due to the nature of this method, it currently takes a moderately long time to run,
    i.e. 7.6 seconds per time unit on AMD Ryzen 5 3600 6-core processor.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range used in original methodology (100-500 hPa & 16.7-58.25 N, 42.5-220.5 E)):
        uv_sub = uv_data.sel(plev=slice(100, 500), lat=slice(16.7, 58.25), lon=slice(42.5, 220.5))

        # Run algorithm:
        schiemann_outputs = jsmetrics.jet_core_algorithms.schiemann_et_al_2009(uv_sub, ws_threshold=30)

        # Get jet occurence for first day in data
        schiemann_jet_occurence_first_day = schiemann_outputs['jet_occurence'].isel(time=0).max('plev')

        # Produce a jet occurence count across all pressure levels
        schiemann_jet_counts_all_levels = schiemann_outputs['jet_occurence'].sum(('time', 'plev'))

        # Use the jet occurence values as a mask to extract the jet windspeeds
        schiemann_jet_ws = schiemann_outputs.where(schiemann_outputs['jet_occurence'] > 0)['ws']
    """
    #  Step 1. Calculate wind vector
    data["ws"] = windspeed_utils.get_resultant_wind(data["ua"], data["va"])

    #  Step 2. Calculate jet occurences
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1:
        if "time" in data.dims:
            data = data.squeeze("time")
        output = jet_core_algorithms_components.get_local_jet_occurence(
            data, ws_threshold=ws_threshold, u_threshold=u_threshold
        )
    else:
        output = data.groupby("time").map(
            jet_core_algorithms_components.get_local_jet_occurence,
            (
                ws_threshold,
                u_threshold,
            ),
        )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def manney_et_al_2011(
    data,
    jet_core_plev_limit,
    jet_core_ws_threshold=40,
    jet_boundary_ws_threshold=30,
    ws_drop_threshold=25,
    jet_core_lat_distance=15,
    check_diagonals=False,
):
    r"""
    This method detects jet cores (within an altitude range see 'jet_core_plev_limit') and a boundary region surrounding
    those cores based on two windspeed thresholds. Two checks are applied after initial detection of cores to check whether
    boundaries with more then one core are part of the same feature (the default threshold for these boundaries is 30 m/s,
    see 'jet_boundary_ws_threshold'). The two checks seperate cores based on whether the cores are more than a certain distance apart
    (default is 15 degrees, see 'jet_core_lat_distance') and whether the windspeed between two given cores does not drop
    below a windspeed threshold (default is 25 m/s, see 'ws_drop_threshold')

    This method returns four outputs
        1. **jet_core_mask** -- Regions within each latitude-altitude slice that are local maxima and have windspeeds above the 'jet_core_ws_threshold'
        2. **jet_region_mask** -- Regions above, below, left and right of any given jet core with windspeed above the 'jet_boundary_ws_threshold'
        3. **jet_region_contour_mask** -- All contigious regions of windspeeds emcompassing a jet core above the 'jet_boundary_ws_threshold' (i.e. not just above, below, left and right)
        4. **ws** -- Resultant wind speed calculated from 'ua', 'va' inputs.

    This method was originally introduce in Manney et al. (2011) (https://doi.org/10.5194/acp-11-6115-2011),
    and is described in Section 3.1 of that study. This method is also known as the JETPAC (Jet and
    Tropopause Products for Analysis and Characterization) software package, and available in its original
    form at request to NASA JPL.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    jet_core_plev_limit: tuple or array
        Sequence of two values relating to the pressure level limit of the jet cores (original paper uses 100-400 hPa)
    jet_core_ws_threshold : int or float
        Threshold used for jet cores (default=40 m/s)
    jet_boundary_ws_threshold : int or float
        Threshold for jet boundaries (default=30 m/s)
    ws_drop_threshold : int or float
        Threshold for drop in windspeed along direct interpolated path between cores (default: 25 m/s)
    jet_core_lat_distance : int or float
        Threshold for maximum distance between cores to be counted the same (default: 15 degrees)
    check_diagonals : bool
        Whether to check the diagonal edges of each maxima in the latitude-altitude plane. The original method does not include this (default: False).

    Returns
    ----------
    output : xarray.Dataset
        Data containing the four output variables: 'ws', 'jet_region_mask', 'jet_region_contour_mask', and 'jet_core_mask'

    Notes
    -----
    The implementation of this method varies slightly from the original, because this method will return a mask rather
    than dynamical values. The intention of this, is to allow these masks to be used to subset other variables such as windspeed
    (see 'Examples' for demonstration of how to use these masks).

    There is an update to this method introduced in Manney & Hegglin 2018 to include physically-based method to extract the
    subtropical jet is identified (and thus distinguished from polar jets). This update is not included in jsmetrics.

    'jet_region_above_ws_threshold_mask' is provided here, but not explicitly in the original methodology. It is meant as
    an alternative to using a contour to check which regions encompass jet cores.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range appropriate for original methodology (100-1000 hPa)):
        uv_sub = uv_data.sel(plev=slice(100, 1900))

        # Run algorithm:
        manney_outputs = jsmetrics.jet_core_algorithms.manney_et_al_2011(uv_sub, ws_core_threshold=40, ws_boundary_threshold=30, jet_core_plev_limit=(100, 400))

        # Use the jet core mask to extract the jet windspeeds
        manney_jet_ws = manney_outputs.where(manney_outputs['jet_core_mask'])['ws']

    """
    # Step 1. Check plev and time coordinate in data
    if "plev" not in data.dims:
        data = data.expand_dims("plev")
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")

    # Step 2. Check a pressure level limit is provided by the user
    if not jet_core_plev_limit:
        raise KeyError(
            "Please provide a pressure level limit for jet cores returned by this metric. As an example the original methodology used a limit of 100-400 hPa. To replicate this, pass the parameter jet_core_plev_limit=(100, 400)."
        )

    # Step 3. Calculate wind speed from ua and va components.
    data["ws"] = windspeed_utils.get_resultant_wind(data["ua"], data["va"])

    # Step 4. Run Algorithm and return outputs
    if data["time"].size == 1:
        if "time" in data.dims:
            data = data.squeeze("time")
        output = (
            jet_core_algorithms_components.run_jet_core_and_region_algorithm_on_one_day(
                data,
                jet_core_plev_limit,
                jet_core_ws_threshold,
                jet_boundary_ws_threshold,
                ws_drop_threshold,
                jet_core_lat_distance,
                check_diagonals,
            )
        )
    else:
        output = data.groupby("time").map(
            jet_core_algorithms_components.run_jet_core_and_region_algorithm_on_one_day,
            (
                jet_core_plev_limit,
                jet_core_ws_threshold,
                jet_boundary_ws_threshold,
                ws_drop_threshold,
                jet_core_lat_distance,
                check_diagonals,
            ),
        )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def penaortiz_et_al_2013(data, ws_threshold=30):
    r"""
    This method follows a two step procedure for calculating local wind maxima and then subcategorising the local maxima into
    two distinct jet masks: the Subtropical Jet (STJ) and Polar Front Jet (PFJ).

    This method returns 4 outputs:
        1. **local_wind_maxima** -- Binary mask of local wind maxima
        2. **local_wind_maxima_by_monthyear** -- Same as above, but for monthyear frequency
        3. **polar_front_jet** -- Binary mask of the PFJ (by monthyear).
        4. **subtropical_jet** -- Binary ask of the STJ (by monthyear).

    This method was first introduced in Pena-Ortiz et al. (2013) (https://doi.org/10.1002/jgrd.50305) and
    is described in section 2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    ws_threshold : int or float
        windspeed threshold to apply (default=30 ms-1)

    Returns
    ----------
    output : xarray.Dataset
        Data containing the four output variables: 'local_wind_maxima', 'local_wind_maxima_by_monthyear', 'polar_front_jet', and 'subtropical_jet'

    Notes
    -----
    See Table 1 in the respective paper for the categories used to seperate the STJ and PFJ.
    The STJ is only seperated in DJF for the Northern Hemisphere.

    **Slow method**: currently takes a long time i.e. 1.3 seconds per time unit with 8 plevs (i.e. 1.3 seconds per day)
    on a AMD Ryzen 5 3600 6-core processor.

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range used in original methodology (100-400 hPa)):
        uv_sub = uv_data.sel(plev=slice(100, 400))

        # Run algorithm:
        pena_outputs = jsmetrics.penaortiz_et_al_2013(uv_sub)

        # Produce a count of polar front jet values across all pressure levels
        pena_pfj_counts_all_levels = pena_outputs['polar_front_jet'].sum(('monthyear', 'plev'))

        # Use the polar front jet values as a mask to extract the jet windspeeds
        uv_sub["ws_by_monthyear"] = (
            data["ws"]
            .resample(time="MS")
            .sum()
            .rename({"time": "monthyear"})
        )
        pena_pfj_ws = pena_outputs.where(pena_outputs['polar_front_jet'] > 0)['ws_by_monthyear']

    """
    #  Step 1. Calculate wind vector
    data["ws"] = windspeed_utils.get_resultant_wind(data["ua"], data["va"])

    #  Step 2. Make array of zeros for local wind maxima location algorithm
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1 and "time" not in data.dims:
        data = data.expand_dims("time")
    output = jet_core_algorithms_components.get_empty_local_wind_maxima_data(data)

    #  Step 3. Find local wind maxima locations by day
    output = output.groupby("time").map(
        jet_core_algorithms_components.get_local_wind_maxima_by_timeunit,
        (ws_threshold,),
    )

    #  Step 4. Get number of days per month with local wind maxima
    output = jet_core_algorithms_components.get_number_of_timeunits_per_monthyear_with_local_wind_maxima(
        output
    )

    #  Step 5. Sort monthyear data into PJ and STJ
    output = jet_core_algorithms_components.subdivide_local_wind_maxima_into_stj_pfj(
        output
    )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def kuang_et_al_2014(data, occurence_ws_threshold=30):
    r"""
    This method produces an event-based jet occurrences and jet center occurrences of the jet stream
    in a given atmospheric column.

    The outputs of this method will produce categorical values of three types:
        0. is not determined to be part of the jet
        1. is a jet occurence
        2. is jet core (upgraded from a jet occurence)

    This method was first introduced in Kuang et al (2014) (https://doi.org/10.1007/s00704-013-0994-x) and
    is described in section 2 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.


    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    occurence_ws_threshold : int or float
        Threshold used to identify a jet-stream occurence point (default=30)

    Returns
    ----------
    output : xarray.Dataset
        Data containing jet-occurence and jet-centres (1 is occurence, 2 is core, 0 is no jet)

    Notes
    -----
    **Slow method**: currently takes a long time i.e. 2 seconds per time unit with 1 plev (i.e. 2 seconds per day) on AMD Ryzen 5 3600 6-core processor

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range used in original methodology (200 hPa)):
        uv_sub = uv_data.sel(plev=slice(200, 200))

        # Run algorithm:
        kuang_outputs = jsmetrics.kuang_et_al_2014(uv_sub, occurence_ws_threshold=30)

        # Extract only jet centers
        kuang_jet_centers = kuang_outputs.where(kuang_outputs['jet_ocurrence1_jet_centre2']==2)

        # Produce a count of jet centers across all pressure levels
        kuang_jet_centers_counts_all_levels = kuang_jet_centers.sum(('time', 'plev'))

        # Use the jet occurence values as a mask to extract the jet windspeeds
        kuang_jet_ws = kuang_outputs.where(kuang_outputs['jet_ocurrence1_jet_centre2'] > 0)['ws']

    """
    # Step 1. Check one and only plev is provided by data input
    if "plev" in data.dims:
        if data["plev"].count() == 1:
            data = data.isel(plev=0)
        else:
            raise ValueError("Please subset to one plev value for algorithm")

    # Step 2. Run Jet-stream Occurence and Centre Algorithm and return outputs
    if "time" not in data.coords:
        output = (
            jet_core_algorithms_components.run_jet_occurence_and_centre_alg_on_one_day(
                data, occurence_ws_threshold
            )
        )
    else:
        if data["time"].size == 1:
            if "time" in data.dims:
                data = data.squeeze("time")
            output = jet_core_algorithms_components.run_jet_occurence_and_centre_alg_on_one_day(
                data, occurence_ws_threshold
            )
        else:
            output = data.groupby("time").map(
                jet_core_algorithms_components.run_jet_occurence_and_centre_alg_on_one_day,
                (occurence_ws_threshold,),
            )
    return output


@sort_xarray_data_coords(coords=["lat", "lon"])
def jet_core_identification_algorithm(
    data, ws_core_threshold=40, ws_boundary_threshold=30
):
    r"""
    This method extract seperate jet cores based on boundary and core windspeed thresholds.

    The output variable of this method includes two types:
        - 0 -- regions not determined to be part of the jet
        - 1-n -- Seperate jet core regions seperated by one condition: if two cores in the same region are more than 15 degrees of latitude away

    This method is inspired by the method from Manney et al. (2011) (https://doi.org/10.5194/acp-11-6115-2011),
    which is described in Section 3.1 of that study.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.
    ws_core_threshold : int or float
        Threshold used for jet-stream core point (default=40)
    ws_boundary_threshold : int or float
        Threshold for jet-stream boundary point (default=30)

    Returns
    ----------
    output : xarray.Dataset
        Data containing the variable 'jet-core_id' (ID number relates to each unique core)

    Notes
    -----
    **Slow method**: currently takes a long time i.e. 7.6 seconds per time unit with 8 plevs
    (i.e. 7.6 seconds per day) on AMD Ryzen 5 3600 6-core processor

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range appropriate for methodology (100-400 hPa)):
        uv_sub = uv_data.sel(plev=slice(100, 400))

        # Run algorithm:
        jca_outputs = jsmetrics.jet_core_algorithms.jet_core_identification_algorithm(uv_sub, ws_core_threshold=40, ws_boundary_threshold=30)

    """
    # Step 1. Check plev coordinate in data
    if "plev" not in data.dims:
        data = data.expand_dims("plev")

    # Step 2. Run Jet-stream Core Idenfication Algorithm and return outputs
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1:
        if "time" in data.dims:
            data = data.squeeze("time")
        output = jet_core_algorithms_components.run_jet_core_algorithm_on_one_day(
            data, ws_core_threshold, ws_boundary_threshold
        )
    else:
        output = data.groupby("time").map(
            jet_core_algorithms_components.run_jet_core_algorithm_on_one_day,
            (
                ws_core_threshold,
                ws_boundary_threshold,
            ),
        )
    return output



from jsmetrics.metrics import jet_core_algorithms, jet_statistics, waviness_metrics


# RULES for METRIC_DETAILS dictionary:
# 1. must have the keys: 'variables', 'coords' 'metric', 'plev_units', 'metric', 'name', 'description, and 'doi'
# 2. 'variables' will contain the required variable names in the data conforming to the CMIP Controlled Vocabulary (Taylor et al. 2011)
# 3. 'coords' will contain the required conforming to the CMIP Controlled Vocabulary
#       and each coord will provide a list of 2 integer values: a minimum and maximum value
#    3.1 for 'plev' coord it is in hPa and higher pressure is considered
#        the minimum value (e.g. 850 hPa - 500 hPa)
# 4. 'plev_units' is units for the pressure level (i.e. hPa, mbar, Pa)
# 5. 'metric' is the relative path to the function
# 6. 'name' is the name of the paper
# 7. 'doi' is the DOI link


METRIC_DETAILS = {
    "Koch2006": {
        "variables": ["ua", "va"],
        "coords": {"plev": [100, 400]},
        "plev_units": "hPa",
        "metric": jet_core_algorithms.koch_et_al_2006,
        "name": "Koch et al. 2006",
        "description": """This a two-step method for detecting \'jet-event occurences\' based on a calculating a weighted average windspeed and then applying a windspeed threshold to isolate jet events from the weighted average. Note that the original method also includes a third step (production of a climatology), but we instead provide an example of how to calculate this third step in the 'Examples' section of the methods docstring. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics""",
        "doi": "https://doi.org/10.1002/joc.1255",
    },
    "ArcherCaldeira2008": {
        "variables": ["ua", "va"],
        "coords": {"plev": [100, 400]},
        "plev_units": "hPa",
        "metric": jet_statistics.archer_caldeira_2008,
        "name": "Archer & Caldeira 2008",
        "description": "This method extracts three monthly-averaged jet stream properties via integrated quantities (windspeed, pressure and latitude) from u-component wind. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics",
        "doi": "https://doi.org/10.1029/2008GL033614",
    },
    "Schiemann2009": {
        "variables": ["ua", "va"],
        "coords": {"plev": [100, 500]},
        "plev_units": "hPa",
        "metric": jet_core_algorithms.schiemann_et_al_2009,
        "name": "Schiemann et al 2009",
        "description": "This method detects jet occurrences, whereby each jet occurence is detected based on three rules applied to inputted wind speed. These rules are based on whether input windspeed is a local maxima (in y and z dimension), and applying various thresholds. Originally the threshold is >= 30 ms for wind vector and >= 0 ms for input u-wind. This method is described in section 2 of that study. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics",
        "doi": "https://doi.org/10.1175/2008JCLI2625.1",
    },
    "Woollings2010_NorthAtlantic": {
        "variables": ["ua"],
        "coords": {"plev": [700, 925], "lat": [15, 75], "lon": [300, 360]},
        "plev_units": "hPa",
        "metric": jet_statistics.woollings_et_al_2010,
        "name": "Woollings et al. 2010 North Atlantic",
        "description": "This method follows an in-text description of 4-steps describing the algorithm of jet-stream identification from Woollings et al. (2010). For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics",
        "doi": "https://onlinelibrary.wiley.com/doi/10.1002/qj.625",
    },
    "Manney2011": {
        "variables": ["ua", "va"],
        "coords": {"plev": [100, 1000]},
        "plev_units": "hPa",
        "metric": jet_core_algorithms.manney_et_al_2011,
        "name": "Manney et al. 2011",
        "description": "This method detects jet cores and defines a boundary region beside those cores based on two windspeed thresholds. Two additional checks are applied after initial detection of cores to check whether cores within the same windspeed region are part of the same feature (see docstring for this method for more information).",
        "doi": "https://doi.org/10.5194/acp-11-6115-2011",
    },
    "BarnesPolvani2013_NorthAtlantic": {
        "variables": ["ua"],
        "coords": {"plev": [700, 850], "lat": [0, 90], "lon": [300, 360]},
        "plev_units": "hPa",
        "metric": jet_statistics.barnes_polvani_2013,
        "name": "Barnes & Polvani 2013 North Atlantic",
        "description": "This method constructs the 'eddy-driven jet' by performing a pressure-weighted average of zonal winds. The winds are then low-pass frequency filtered at each grid point using a 10-day Lanczos filter with 41 weights by default. Finally a 0.01 degree quadratic function is fitted to the peak of the subsequent wind speed profile for each time step. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-12-00536.1",
    },
    "BarnesPolvani2013_NorthPacific": {
        "variables": ["ua"],
        "coords": {"plev": [700, 850], "lat": [0, 90], "lon": [135, 235]},
        "plev_units": "hPa",
        "metric": jet_statistics.barnes_polvani_2013,
        "name": "Barnes & Polvani 2013 North Pacific",
        "description": "This method constructs the 'eddy-driven jet' by performing a pressure-weighted average of zonal winds. The winds are then low-pass frequency filtered at each grid point using a 10-day Lanczos filter with 41 weights by default. Finally a 0.01 degree quadratic function is fitted to the peak of the subsequent wind speed profile for each time step. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-12-00536.1",
    },
    "BarnesPolvani2013_SouthernHemisphere": {
        "variables": ["ua"],
        "coords": {"plev": [700, 850], "lat": [-90, 0]},
        "plev_units": "hPa",
        "metric": jet_statistics.barnes_polvani_2013,
        "name": "Barnes & Polvani 2013 Southern Hemisphere",
        "description": "This method constructs the 'eddy-driven jet' by performing a pressure-weighted average of zonal winds. The winds are then low-pass frequency filtered at each grid point using a 10-day Lanczos filter with 41 weights by default. Finally a 0.01 degree quadratic function is fitted to the peak of the subsequent wind speed profile for each time step. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-12-00536.1",
    },
    "PenaOrtiz2013": {
        "variables": ["ua", "va"],
        "coords": {"plev": [100, 400]},
        "plev_units": "hPa",
        "metric": jet_core_algorithms.penaortiz_et_al_2013,
        "name": "Pena-Ortiz et al. 2013",
        "description": "This method follows a two step procedure for calculating local wind maxima and then subcategorising the local maxima into two distinct jet masks: the Subtropical Jet (STJ) and Polar Front Jet (PFJ). For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1002/jgrd.50305",
    },
    "Kuang2014": {
        "variables": ["ua", "va"],
        "coords": {"plev": [200, 200]},
        "plev_units": "hPa",
        "metric": jet_core_algorithms.kuang_et_al_2014,
        "name": "Kuang et al. 2014.",
        "description": "This method produces an event-based jet occurrences and jet center occurrences of the jet stream in a given atmospheric column. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1007/s00704-013-0994-x",
    },
    "BarnesPolvani2015_NorthAmerica_NorthAtlantic": {
        "variables": ["ua"],
        "coords": {"plev": [700, 925], "lat": [30, 70], "lon": [230, 350]},
        "plev_units": "hPa",
        "metric": jet_statistics.barnes_polvani_2015,
        "name": "Barnes & Polvani 2015",
        "description": "This method calculates the jet positon and wind speed at that position by fitting a parabola around the maximum of zonally average u-component wind speed, using the magnitude at the maximum ('jet_speed') and latitude at that maximum ('jet_lat'). For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-14-00589.1",
    },
    "FrancisVavrus2015": {
        "variables": ["ua", "va"],
        "coords": {"plev": [500, 500], "lat": [20, 80]},
        "plev_units": "hPa",
        "metric": waviness_metrics.francis_vavrus_2015,
        "name": "Francis & Vavrus 2015",
        "description": """This method calculates a waviness metric called: Meridional Circulation Index (MCI) from u- and v-components of wind. For more information see 'Notes' for this metho available on the ReadTheDocs for jsmetrics.""",
        "doi": "https://doi.org/10.1088/1748-9326/10/1/014005",
    },
    "Cattiaux2016": {
        "variables": ["zg"],
        "coords": {"plev": [500, 500]},
        "plev_units": "hPa",
        "metric": waviness_metrics.cattiaux_et_al_2016,
        "name": "Cattiaux et al. 2016",
        "description": "This method calculates a sinousity metric for upper-air flow using geopotential height. This is achieved by comparing an isohype at the Z500 average over 30-70 N to the perimeter at 50∘N. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1002/2016GL070309",
    },
    "GrisePolvani2014": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [-65, -30]},
        "plev_units": "hPa",
        "metric": jet_statistics.grise_polvani_2014,
        "name": "Grise & Polvani 2014",
        "description": "This method calculates the latitude of the midlatitude eddy-driven jet by finding the peak value of the input u-wind field. A polynomial fit is then applied to get an appropriate value of 'jet_lat' at a resolution 0.01 degrees. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1002/2013GL058466",
    },
    "BarnesSimpson2017_NorthAtlantic": {
        "variables": ["ua"],
        "coords": {"plev": [700, 700], "lat": [1, 90], "lon": [280, 350]},
        "plev_units": "hPa",
        "metric": jet_statistics.barnes_simpson_2017,
        "name": "Barnes & Simpson 2017 North Atlantic",
        "description": "This method defines two outputs: 'jet_lat' and 'jet_speed' which are defined as the latitude and speed of the 10-day-averaged maximum zonally-averaged wind speed. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-17-0299.1",
    },
    "BarnesSimpson2017_NorthPacific": {
        "variables": ["ua"],
        "coords": {"plev": [700, 700], "lat": [1, 90], "lon": [120, 230]},
        "plev_units": "hPa",
        "metric": jet_statistics.barnes_simpson_2017,
        "name": "Barnes & Simpson 2017 North Pacific",
        "description": "This method defines two outputs: 'jet_lat' and 'jet_speed' which are defined as the latitude and speed of the 10-day-averaged maximum zonally-averaged wind speed. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-17-0299.1",
    },
    "Bracegirdle2018_SouthernHemisphere": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [-75, -10]},
        "plev_units": "hPa",
        "metric": jet_statistics.bracegirdle_et_al_2018,
        "name": " Bracegirdle et al. 2018",
        "description": "This method calculates the seasonal and annual jet-stream position ('JPOS') and strength ('JSTR') by applying a 0.075 degree cubic spline interpolation to zonally-averaged wind climatology and selecting the maximum. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "https://doi.org/10.1175/JCLI-D-17-0320.1",
    },
    "Ceppi2018_NorthAtlantic_Europe": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [30, 60], "lon": [300, 60]},
        "plev_units": "hPa",
        "metric": jet_statistics.ceppi_et_al_2018,
        "name": "Ceppi et al. 2018 North Atlantic",
        "description": "This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "10.1175/JCLI-D-17-0323.1",
    },
    "Ceppi2018_NorthPacific": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [30, 60], "lon": [140, 240]},
        "plev_units": "hPa",
        "metric": jet_statistics.ceppi_et_al_2018,
        "name": "Ceppi et al. 2018 North Pacific",
        "description": "This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "10.1175/JCLI-D-17-0323.1",
    },
    "Ceppi2018_SouthernHemisphere": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [-60, -30]},
        "plev_units": "hPa",
        "metric": jet_statistics.ceppi_et_al_2018,
        "name": "Ceppi et al. 2018 Southern Hemisphere",
        "description": "This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "10.1175/JCLI-D-17-0323.1",
    },
    "Zappa2018_NorthAtlantic": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [20, 70], "lon": [300, 360]},
        "plev_units": "hPa",
        "metric": jet_statistics.zappa_et_al_2018,
        "name": "Zappa et al. 2018 North Atlantic",
        "description": "This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile. It differs from Ceppi et al. 2018 by also flooring the u-wind input to 0. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "10.1002/2017GL076096",
    },
    "Zappa2018_NorthPacific": {
        "variables": ["ua"],
        "coords": {"plev": [850, 850], "lat": [20, 70], "lon": [140, 240]},
        "plev_units": "hPa",
        "metric": jet_statistics.zappa_et_al_2018,
        "name": "Zappa et al. 2018 North Pacific",
        "description": "This method calculates the jet latitude ('jet-lat') as defined by selecting the centroid of a zonally-averaged wind profile. It differs from Ceppi et al. 2018 by also flooring the u-wind input to 0. For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "10.1002/2017GL076096",
    },
    "Kerr2020_NorthernHemisphere": {
        "variables": ["ua"],
        "coords": {"plev": [500, 500], "lat": [20, 70]},
        "plev_units": "hPa",
        "metric": jet_statistics.kerr_et_al_2020,
        "name": "Kerr et al. 2020",
        "description": "This method defines the latitude and speed of the jet-stream where the maximum zonal winds occur for each longitude and for each time unit (i.e. day). These values are then smoothed across the longitudes with a rectangular pulse (by default this has a width of 10 degrees). For more information see 'Notes' for this method available on the ReadTheDocs for jsmetrics.",
        "doi": "10.1029/2020JD032735",
    },
}



# -*- coding: utf-8 -*-

"""
    Statistics and algorithms for determining the "waviness" of upper-level mean flow within a given
    time window. These metrics only have meaning at an integrated global scale.

    The following metrics each return a xarray.Dataset and are ordered by paper publish year.
"""

# imports
import xarray
from . import waviness_metrics_components
from jsmetrics.utils import spatial_utils
from jsmetrics.core.check_data import sort_xarray_data_coords

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


@sort_xarray_data_coords(coords=["lat", "lon"])
def francis_vavrus_2015(data):
    r"""
    This method calculates a waviness metric: Meridional Circulation Index (MCI) from u- and v-components of wind.

    MCI is calculated as follows:

    .. math::
        MCI = \frac{v*|v|}{u^2+v^2}

    When MCI = 0, the wind is purely zonal, and when MCI= 1 (-1), the flow is from the South (North).

    This method was originally introduce in Francis & Vavrus (2015) https://doi.org/10.1088/1748-9326/10/1/014005
    and is described in the Section entitled: 'Meridional circulation Index (MCI)'.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'ua' and 'va', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the one output: 'mci'

    Notes
    -----
    In the original methodology, MCI is expressed in terms of seasonal anomaly, we show you how to do this in 'Examples'

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        uv_data = xr.open_dataset('path_to_uv_data')

        # Subset dataset to range used in original methodology (500 hPa & 20-80 N)):
        uv_sub = uv_data.sel(plev=500, lat=slice(20, 80))

        # Run statistic:
        fv15 = jsmetrics.waviness_metrics.francis_vavrus_2015(uv_sub)

        # Express MCI as a seasonal anomaly
        fv15_seasonal_anomalies = (mci['mci'] - mci['mci'].groupby('time.season').mean())
    """
    #  Step 1. calculating Meridional Circulation Index from data
    data["mci"] = waviness_metrics_components.calc_meridional_circulation_index(data)
    return data


@sort_xarray_data_coords(coords=["lat", "lon"])
def cattiaux_et_al_2016(data):
    r"""
    This method calculates a sinousity metric for upper-air flow using geopotential height.
    The value of sinuosity is selected using an isohypse which precisely corresponds to the Z500 average over 30-70∘N .
    Then this value is compared to the perimeter at 50∘N to calculate sinuosity.

    This method was first introduce in Cattiaux et al (2016) https://doi.org/10.1002/2016GL070309 and
    is described in section 3.1 of that study.

    Please see 'Notes' below for any additional information about the implementation of this method
    to this package.

    Parameters
    ----------
    data : xarray.Dataset
        Data which should containing the variables: 'zg', and the coordinates: 'lon', 'lat', 'plev' and 'time'.

    Returns
    ----------
    output : xarray.Dataset
        Data containing the two outputs: 'sinousity' and 'zonal_mean_zg_30Nto70N'

    Notes
    -----
    The original implementation used the R package 'geosphere' to calculate sinuosity. Here we use the Python package
    Shapely to calculate great circle distances between the two perimeters to calculate sinuosity.

    **Moderately slow method:** currently takes a moderate amount of time i.e. 2 seconds per 100 time units\
    with 1 plev on AMD Ryzen 5 3600 6-core processor

    Examples
    --------
    .. code-block:: python

        import jsmetrics
        import xarray as xr

        # Load in dataset with u and v components:
        zg_data = xr.open_dataset('path_to_zg_data')

        # Subset dataset to range used in original methodology (500 hPa & 00-90 N)):
        zg_sub = zg_data.sel(plev=500, lat=slice(0, 90))

        # Run statistic:
        c16 = jsmetrics.waviness_metrics.cattiaux_et_al_2016(zg_sub)
    """
    # Check input is xarray.Dataset
    if isinstance(data, xarray.DataArray):
        data = data.to_dataset()

    #  Step 1. get zonal average for each timestep
    if "time" not in data.coords:
        raise KeyError("Please provide a time coordinate for data to run this metric")
    if data["time"].size == 1:
        data = data.expand_dims("time")
    data["zonal_mean_zg_30Nto70N"] = (
        data["zg"].sel(lat=slice(30, 70)).groupby("time").mean(...)
    )

    #  Step 2. Get latitude circle of 50 N
    circle_50N = spatial_utils.get_latitude_circle_linestring(50, 0, 360)

    #  Step 3. Loop over each time step and calculate sinuosity
    if data["time"].size == 1:
        # shrink dims again
        data = data.isel(time=0)
        output = waviness_metrics_components.get_sinuosity_of_zonal_mean_zg(
            data, circle_50N
        )
    else:
        output = data.groupby("time").map(
            lambda row: waviness_metrics_components.get_sinuosity_of_zonal_mean_zg(
                row, circle_50N
            )
        )
    return output



# -*- coding: utf-8 -*-

"""
    Various utility functions needed for the jet-stream metrics and algorithms not belonging to windspeed or spatial utils.
    The module is built from xarray data-structures, so this file contains all the stuff that helps the library handle xarray dataset/dataarrays

    Classes and Functions ordered alphabetically
"""

# imports
import cftime
import itertools
import numpy as np
import xarray as xr
import scipy.signal

# docs
__author__ = "Thomas Keel"
__email__ = "thomas.keel.18@ucl.ac.uk"
__status__ = "Development"


def add_num_of_days_to_360Datetime(datetime_360day, num_of_days_to_add):
    """
    Adds a number of days to cftime.Datetime360Day format

    Parameters
    ----------
    datetime_360day : cftime.Datetime360Day
        Date to add days to

    num_of_days_to_add : int or float
        Number of days to add to 360 day format

    Returns
    ----------
    new_360day_date : cftime.Datetime360Day
        Date with added number of days

    Raises
    ----------
    AssertionError :
        If not '360_day' format or not a cftime.Datetime object
    ValueError :
        If number of days not more than 0
    """
    assert hasattr(
        datetime_360day, "calendar"
    ), "date type inputted is not in cftime format."
    assert (
        getattr(datetime_360day, "calendar") == "360_day"
    ), "input date is not '360_day' format."
    if num_of_days_to_add <= 0:
        raise ValueError("Number of days to add to date is not more than 0")

    new_day = ((datetime_360day.day + num_of_days_to_add) - 30) % 30
    if new_day == 0:
        new_day = 30
    num_of_months_to_add = (datetime_360day.day + num_of_days_to_add) / 30
    if num_of_months_to_add % 1 == 0:
        num_of_months_to_add = num_of_months_to_add - 1
    else:
        num_of_months_to_add = np.floor(num_of_months_to_add)
    new_month = ((datetime_360day.month + num_of_months_to_add) - 12) % 12
    if new_month == 0:
        new_month = 12
    num_of_years_to_add = (datetime_360day.month + num_of_months_to_add) / 12
    if num_of_years_to_add % 1 == 0:
        num_of_years_to_add = num_of_years_to_add - 1
    else:
        num_of_years_to_add = np.floor(num_of_years_to_add)

    new_year = datetime_360day.year + num_of_years_to_add
    new_360day_date = cftime.Datetime360Day(
        day=new_day,
        month=new_month,
        year=new_year,
        hour=datetime_360day.hour,
        minute=datetime_360day.minute,
        second=datetime_360day.second,
    )
    return new_360day_date


def add_num_of_days_to_NoLeapDatetime(datetime_noleap, num_of_days_to_add):
    """
    Adds a number of days to cftime.DatetimeNoLeap format

    Parameters
    ----------
    datetime_noleap : cftime.DatetimeNoLeap
        Date to add days to

    num_of_days_to_add : int or float
        Number of days to add to noleap day format

    Returns
    ----------
    new_noleap_date : cftime.DatetimeNoLeap
        Date with added number of days

    Raises
    ----------
    AssertionError :
        If not 'noleap' format or not a cftime.Datetime object
    ValueError :
        If number of days not more than 0
    """
    assert hasattr(
        datetime_noleap, "calendar"
    ), "date type inputted is not in cftime format."
    assert (
        getattr(datetime_noleap, "calendar") == "noleap"
    ), "input date is not 'no_leap' format."
    if num_of_days_to_add <= 0:
        raise ValueError("Number of days to add to date is not more than 0")

    return cftime.DatetimeNoLeap.fromordinal(
        cftime.DatetimeNoLeap.toordinal(datetime_noleap) + num_of_days_to_add,
        calendar="noleap",
        has_year_zero=True,
    )


def add_pad_to_array(arr, pad_width=1, constant_values=0):
    """
    Add a edge of constant values around a numpy array

    Parameters
    ----------
    arr : array-like
        A 2-D array with numeric values (i.e. dtypes float or int)
    pad_width : int
        Width of values to add to edges (default: 1)
    constant_values : int
        Value to add to edges (default: 0)

    Returns
    ----------
    padded_arr : array-like
        A 2-D array with new dimensions as a pad has been added at the edges
    """
    padded_arr = np.pad(
        arr, pad_width, mode="constant", constant_values=constant_values
    )
    return padded_arr


def check_at_least_n_plevs_in_data(data, n_plevs):
    """
    Checks there are at least two pressure-levels (plevs) in xarray dataset

    Parameters
    ----------
    data : xarray.Dataset
        Data to check

    Raises
    ----------
    ValueError :
        If not two pressure levels
    """
    if data["plev"].count() < n_plevs:
        raise ValueError(
            "Need at least %s pressure levels (plevs) for calculation" % (n_plevs)
        )


def check_coords_in_data(data, req_coords):
    """
    Will check if a given coordinates are in the data and raises a key error if any of them are not.
    This function is needed to check before algorithm continues and uses too much memory.
    Built from xarray.

    Parameters
    ----------
    data : xarray.Dataset
        Data to check
    req_coords : array-like or tuple
        Coordinates to check if in data

    Raises
    ----------
    KeyError :
        If any given coord is not in data
    """
    for coord in req_coords:
        if coord not in data.coords:
            raise KeyError("'%s' is not in the data" % (coord,))


def check_if_data_is_xarray_datatype(data):
    """
    What it says on the tin.

    Parameters
    ----------
    data : xarray.Dataset
        Data to check

    Raises
    ----------
    TypeError :
        If input is not an xarray data type
    """
    if not isinstance(data, xr.Dataset) or isinstance(data, xr.DataArray):
        raise TypeError("input needs to be xarray.DataSet or xarray.DataArray")


def check_plev_units(data, expected_plev_units):
    """
    Will check that pressure level units are in data plev coord (i.e. Pa, hPa, mbar, millibars).

    Parameters
    ----------
    data : xr.DataArray or xr.Dataset
        Data containing plev coord and units for plev
    expected_plev_units : list
        List of names of plev units (i.e. mbar, Pa, hPa, etc.)

    Returns
    ----------
    units : str
        Units of plev coord

    Raises
    ----------
    KeyError
        If plev is not a coord in input 'data'. And if the units of plev are not in expected plev units

    """
    if "plev" not in data.coords:
        raise KeyError(
            "Please provide a plev coordinate for data to determine plev units"
        )
    if not hasattr(data["plev"], "units"):
        raise KeyError(
            f"You will need assign units (i.e {expected_plev_units}) to plev to run this method."
        )

    if data["plev"].units not in expected_plev_units:
        raise KeyError(
            f"plev unit: {data['plev'].units} is not in {expected_plev_units}. You will need assign units to plev to run this method."
        )
    else:
        return data["plev"].units


def check_variables_in_data(data, req_variables):
    """
    What it says on the tin.
    Built from xarray

    Parameters
    ----------
    data : xarray.Dataset
        Data to check
    req_variables : array-like
        Variables needed in data

    Raises
    ----------
    KeyError :
        If variables not in data
    """
    for var in req_variables:
        if var not in data.variables:
            raise KeyError("'%s' is not the data" % (var,))


def find_nearest_value_to_array(array, value):
    """
    Will find the nearest value to a given array
    Built for use with xarray
    adapted from: https://stackoverflow.com/questions/2566412/find-nearest-value-in-numpy-array

    Parameters
    ----------
    array : array-like
        Array to reference
    value : numeric-like
        Number to check nearest value in input array

    Returns
    ----------
    output : numeric
        Closest value to input value in input array
    """
    array = np.asarray(array)
    idx = np.argmin(np.abs(array - value))
    return array[idx]


def get_local_maxima(arr, axis=0):
    """
    Uses scipy.signal.argrelextrema to get index location of minimum value in array

    Taken from
    https://stackoverflow.com/questions/4624970/finding-local-maxima-minima-with-numpy-in-a-1d-numpy-array

    Parameters
    ----------
    arr : array-like
        array to find index location of minima value from
    axis : int
        axis for scipy.signal.argrelextrema
    """
    return scipy.signal.argrelextrema(arr, np.greater, axis=axis)


def get_local_minima(arr, axis=0):
    """
    Uses scipy.signal.argrelextrema to get index location of maximum value in array

    Taken from
    https://stackoverflow.com/questions/4624970/finding-local-maxima-minima-with-numpy-in-a-1d-numpy-array

    Parameters
    ----------
    arr : array-like
        array to find index location of maxima value from
    axis : int
        axis for scipy.signal.argrelextrema
    """
    return scipy.signal.argrelextrema(arr, np.less, axis=axis)


def filter_local_extremes_to_min_distance(local_extrema, min_distance_threshold=2):
    """
    Filter local extremes (i.e. outputs of minima or maxima from scipy.signal.argrelextrema)
    so that no neighbours remain in array.

    Parameters
    ----------
    local_extrema : array-like
        likely (stacked) outputs of scipy.signal.argrelextrema
    min_distance_threshold : int
        Minimum distances between indexes in array (default: 2 indexes)

    Returns
    -------
    filtered_extrema : array-like
        (stacked) outputs of scipy.signal.argrelextrema with values less than a min_distance_threshold away removed

    Examples
    --------
    maxima_indices = np.column_stack(jsmetrics.utils.data_utils.get_local_maxima(current['ws'].data))
    filtered_indices = filter_local_extremes_so_no_neighbours(maxima_indices, min_distance_threshold=2)
    """
    filtered_extrema = []

    for idx in local_extrema:
        if not any(
            np.linalg.norm(idx - existing_maxima) < min_distance_threshold
            for existing_maxima in filtered_extrema
        ):
            filtered_extrema.append(idx)

    filtered_extrema = np.array(filtered_extrema)
    return filtered_extrema


def find_intersection_between_two_array_of_arrays(array1, array2):
    """
    Find the intersection between two arrays of arrays. See examples for example.

    Parameters
    ----------
    array1 : np.array
        First array of arrays to compare with array2
    array2 : np.array
        Second array of arrays to compare with array1

    Returns
    -------
    intersection : np.array
        Intesection of the two arrays returning only the arrays within the original two that are in both.

    Examples
    --------
    array1 = [[1, 2], [3, 4]]
    array2 = [[1, 2], [4, 3]]
    find_intersection_between_two_array_of_arrays(array1, array2)
    # returns [[1, 2]]

    """
    intersection = []
    for arr1 in array1:
        for arr2 in array2:
            if np.array_equal(arr1, arr2):
                intersection.append(arr1)
                break
    intersection = np.array(intersection)
    return intersection


def get_num_of_decimal_places(num):
    """
    Gets number of decimal places in a float

    Parameters
    ----------
    num : float or int
        input number to get decimal places from

    Returns
    ----------
    decimal_places : int
        number of decimal places
    """
    num = "{:f}".format(num).rstrip("0")
    decimal_places = num[::-1].find(".")
    return decimal_places


def remove_duplicates(arr):
    """
    Removes duplicates from array. From:
    https://stackoverflow.com/questions/2213923/removing-duplicates-from-a-list-of-lists

    Parameters
    ----------
    arr : array-like
        arr to remove duplicates from

    Returns
    ----------
    arr : array-like
        arr with no duplicates
    """

    arr.sort()
    return list(v for v, _ in itertools.groupby(arr))


def remove_unwanted_coords_from_data(
    data, wanted_coords, unwanted_coords=(), show_error=False
):
    """
    What it says on the tin.
    Built from xarray

    Parameters
    ----------
    data : xarray.Dataset
        Data to check

    wanted_coords : array-like or tuple
        Coords to retain in data
    unwanted_coords : array-like or tuple
        Coords to remove from data


    Raises
    ----------
    ValueError :
        If coord cannot be removed from data
    """
    dims = set(data.dims)
    for rem in unwanted_coords:
        try:
            dims.remove(rem)
        except Exception as e:
            # a little sloppy, but probably okay
            if show_error:
                print(e)

    wanted_coords = set(wanted_coords)
    difference = dims.difference(set(wanted_coords))
    if len(difference) > 0:
        raise ValueError(
            "Unwanted coords in data: %s.\
             Please subset/remove so that the slice can be 2D."
            % (difference,)
        )


def rescale_lat_resolution(lats, lat_resolution):
    """
    Rescale latitude resolution to input lat resolution in degrees.

    Component of method from Barnes & Polvani (2013) https://doi.org/10.1175/JCLI-D-12-00536.1
    & Grise & Polvani 2014 https://doi.org/10.1002/2013GL058466
    & Bracegirdle et al (2018) https://doi.org/10.1175/JCLI-D-17-0320.1

    Parameters
    ----------
    lats : xr.DataArray or array-like
        Array of latitude values
    lat_resolution : int or float
        Latitude resolution in degrees

    Returns
    ----------
    output : numpy.array
        Rescaled array of latitude values
    """
    return np.arange(min(lats), max(lats) + lat_resolution, lat_resolution)


def slice_array_by_index_breaks(array_to_slice, index_breaks):
    """
    Will break an array down into segments based on a list of indexes containing information about where to create slices

    Parameters
    ----------
    array_to_slice : array-like
        Array to break down based on slice breaks
    index_breaks : array-like
        Indexes from which to slice array

    Returns
    ----------
    output: list
        Broken down slices of original array_to_slice
    """
    return [
        array_to_slice[i:j] for i, j in zip([0] + index_breaks, index_breaks + [None])
    ]
